@@ -26,6 +26,8 @@
   Published with a permission.
 */
 
+#define VER "1.0"
+
 #include <my_global.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -48,8 +50,8 @@ The parts not included are excluded by #ifndef UNIV_INNOCHECKSUM. */
 #include "buf0buf.h"             /* buf_page_is_corrupted */
 #include "page0zip.h"            /* page_zip_*() */
 #include "trx0undo.h"            /* TRX_* */
-#include "ut0crc32.h"		 /* ut_crc32_init() */
 #include "fil0crypt.h"           /* fil_space_verify_crypt_checksum */
+
 #include <string.h>
 
 #ifdef UNIV_NONINL
@@ -77,9 +79,7 @@ static ulint physical_page_size;  /* Page size in bytes on disk. */
 static ulint extent_size;
 static ulint xdes_size;
 ulong srv_page_size;
-ulong srv_page_size_shift;
-static uint32_t dblwr_1;
-static uint32_t dblwr_2;
+uint32_t srv_page_size_shift;
 /* Current page number (0 based). */
 uint32_t		cur_page_num;
 /* Current space. */
@@ -103,10 +103,8 @@ FILE*				log_file = NULL;
 /* Enabled for log write option. */
 static bool			is_log_enabled = false;
 static bool			skip_freed_pages;
-static uint32_t 		tablespace_flags= 0;
 static byte field_ref_zero_buf[UNIV_PAGE_SIZE_MAX];
 const byte *field_ref_zero = field_ref_zero_buf;
-constexpr uint32_t USE_FSP_FLAGS{UINT32_MAX};
 
 #ifndef _WIN32
 /* advisory lock for non-window system. */
@@ -261,11 +259,14 @@ void print_leaf_stats(
 }
 
 /** Init the page size for the tablespace.
-@param[in]	flags	InnoDB tablespace flags */
-static void init_page_size_from_flags(const uint32_t flags)
+@param[in]	buf	buffer used to read the page */
+static void init_page_size(const byte* buf)
 {
+	const unsigned	flags = mach_read_from_4(buf + FIL_PAGE_DATA
+						 + FSP_SPACE_FLAGS);
+
 	if (fil_space_t::full_crc32(flags)) {
-		const ulong ssize = FSP_FLAGS_FCRC32_GET_PAGE_SSIZE(flags);
+		const uint32_t ssize = FSP_FLAGS_FCRC32_GET_PAGE_SSIZE(flags);
 		srv_page_size_shift = UNIV_ZIP_SIZE_SHIFT_MIN - 1 + ssize;
 		srv_page_size = 512U << ssize;
 		physical_page_size = srv_page_size;
@@ -274,7 +275,7 @@ static void init_page_size_from_flags(const uint32_t flags)
 		return;
 	}
 
-	const ulong	ssize = FSP_FLAGS_GET_PAGE_SSIZE(flags);
+	const uint32_t ssize = FSP_FLAGS_GET_PAGE_SSIZE(flags);
 
 	srv_page_size_shift = ssize
 		? UNIV_ZIP_SIZE_SHIFT_MIN - 1 + ssize
@@ -440,12 +441,7 @@ static bool is_page_all_zeroes(
 				with crypt_scheme encrypted
 @param[in]	flags		tablespace flags
 @retval true if page is corrupted otherwise false. */
-static
-bool
-is_page_corrupted(
-	byte*		buf,
-	bool		is_encrypted,
-	ulint		flags)
+static bool is_page_corrupted(byte *buf, bool is_encrypted, uint32_t flags)
 {
 
 	/* enable if page is corrupted. */
@@ -550,15 +546,24 @@ is_page_corrupted(
 	return(is_corrupted);
 }
 
-/** Check if page is doublewrite buffer or not.
-@retval true  if page is doublewrite buffer otherwise false. */
-static bool is_page_doublewritebuffer()
+/********************************************//*
+ Check if page is doublewrite buffer or not.
+ @param [in] page	buffer page
+
+ @retval true  if page is doublewrite buffer otherwise false.
+*/
+static
+bool
+is_page_doublewritebuffer(
+	const byte*	page)
 {
-  if (cur_space != 0) return false;
-  const uint32_t extent{static_cast<uint32_t>(
-    cur_page_num & ~(extent_size - 1))};
-  return cur_page_num > FSP_DICT_HDR_PAGE_NO &&
-	 extent && (extent == dblwr_1 || extent == dblwr_2);
+	if ((cur_page_num >= extent_size)
+		&& (cur_page_num < extent_size * 3)) {
+		/* page is doublewrite buffer. */
+		return (true);
+	}
+
+	return (false);
 }
 
 /*******************************************************//*
@@ -593,7 +598,7 @@ Rewrite the checksum for the page.
 @retval false : skip the rewrite as checksum stored match with
 		calculated or page is doublwrite buffer.
 */
-static bool update_checksum(byte* page, ulint flags)
+static bool update_checksum(byte* page, uint32_t flags)
 {
 	ib_uint32_t	checksum = 0;
 	byte		stored1[4];	/* get FIL_PAGE_SPACE_OR_CHKSUM field checksum */
@@ -637,7 +642,7 @@ static bool update_checksum(byte* page, ulint flags)
 	} else if (use_full_crc32) {
 		ulint payload = buf_page_full_crc32_size(page, NULL, NULL)
 			- FIL_PAGE_FCRC32_CHECKSUM;
-		checksum = ut_crc32(page, payload);
+		checksum = my_crc32c(0, page, payload);
 		byte* c = page + payload;
 		if (mach_read_from_4(c) == checksum) return false;
 		mach_write_to_4(c, checksum);
@@ -702,7 +707,7 @@ write_file(
 	const char*	filename,
 	FILE*		file,
 	byte*		buf,
-	ulint		flags,
+	uint32_t	flags,
 	fpos_t*		pos)
 {
 	bool	do_update;
@@ -765,7 +770,7 @@ Parse the page and collect/dump the information about page type
 @param [in] file	file for diagnosis.
 @param [in] is_encrypted  tablespace is encrypted
 */
-static void
+void
 parse_page(
 	const byte*	page,
 	byte*		xdes,
@@ -785,12 +790,6 @@ parse_page(
 	str = skip_page ? "Double_write_buffer" : "-";
 	page_no = mach_read_from_4(page + FIL_PAGE_OFFSET);
 	if (skip_freed_pages) {
-
-		/** Skip doublewrite pages when -r is enabled */
-		if (is_page_doublewritebuffer()) {
-			return;
-		}
-
 		const byte *des= xdes + XDES_ARR_OFFSET +
 			xdes_size * ((page_no & (physical_page_size - 1))
 				     / extent_size);
@@ -985,18 +984,6 @@ parse_page(
 			fprintf(file, "#::" UINT32PF "\t\t|\t\tTransaction system "
 				"page\t\t|\t%s\n", cur_page_num, str);
 		}
-
-		if (cur_space == 0 &&
-		    (mach_read_from_4(page + TRX_SYS_DOUBLEWRITE +
-			             TRX_SYS_DOUBLEWRITE_MAGIC) ==
-		     TRX_SYS_DOUBLEWRITE_MAGIC_N))  {
-			dblwr_1 = mach_read_from_4(
-					page + TRX_SYS_DOUBLEWRITE +
-					TRX_SYS_DOUBLEWRITE_BLOCK1);
-			dblwr_2 = mach_read_from_4(
-					page + TRX_SYS_DOUBLEWRITE +
-					TRX_SYS_DOUBLEWRITE_BLOCK2);
-		}
 		break;
 
 	case FIL_PAGE_TYPE_FSP_HDR:
@@ -1239,27 +1226,10 @@ static struct my_option innochecksum_options[] = {
   {"skip-freed-pages", 'r', "skip freed pages for the tablespace",
    &skip_freed_pages, &skip_freed_pages, 0, GET_BOOL, NO_ARG,
    0, 0, 0, 0, 0, 0},
-  {"tablespace-flags", 0, "InnoDB tablespace flags (default: 4294967295 "
-   "= read from page 0)", &tablespace_flags, &tablespace_flags, 0,
-   GET_UINT, REQUIRED_ARG, USE_FSP_FLAGS, 0, USE_FSP_FLAGS, 0, 0, 0},
 
   {0, 0, 0, 0, 0, 0, GET_NO_ARG, NO_ARG, 0, 0, 0, 0, 0, 0}
 };
 
-/* Print out the Innodb version and machine information. */
-static void print_version(void)
-{
-#ifdef DBUG_OFF
-	printf("%s Ver %s, for %s (%s)\n",
-		my_progname, INNODB_VERSION_STR,
-		SYSTEM_TYPE, MACHINE_TYPE);
-#else
-	printf("%s-debug Ver %s, for %s (%s)\n",
-		my_progname, INNODB_VERSION_STR,
-		SYSTEM_TYPE, MACHINE_TYPE);
-#endif /* DBUG_OFF */
-}
-
 static void usage(void)
 {
 	print_version();
@@ -1316,14 +1286,6 @@ innochecksum_get_one_option(
 		my_end(0);
 		exit(EXIT_SUCCESS);
 		break;
-	default:
-	  if (tablespace_flags != USE_FSP_FLAGS &&
-	      !fil_space_t::is_valid_flags(tablespace_flags, false) &&
-	      !fil_space_t::is_valid_flags(tablespace_flags, true)) {
-		fprintf(stderr, "Error: Provided --tablespace-flags "
-			"is not valid.");
-		return true;
-	  }
 	}
 
 	return(false);
@@ -1403,7 +1365,7 @@ static int verify_checksum(
 	byte*			buf,
 	bool			is_encrypted,
 	unsigned long long*	mismatch_count,
-	ulint			flags)
+	uint32_t		flags)
 {
 	int exit_status = 0;
 	if (is_page_corrupted(buf, is_encrypted, flags)) {
@@ -1444,7 +1406,7 @@ rewrite_checksum(
 	byte*		buf,
 	fpos_t*		pos,
 	bool		is_encrypted,
-	ulint		flags)
+	uint32_t	flags)
 {
 	bool is_compressed = fil_space_t::is_compressed(flags);
 
@@ -1454,87 +1416,6 @@ rewrite_checksum(
 		&& !write_file(filename, fil_in, buf, flags, pos);
 }
 
-/** Read and validate page 0, then initialize tablespace flags
-and page size.
-@param	fil_in        File pointer
-@param	buf           Buffer to read page into
-@return whether the page was read successfully */
-static bool read_and_validate_page0(FILE *fil_in, byte *buf)
-{
-  /* Read the minimum page size first */
-  size_t initial_page_size= UNIV_ZIP_SIZE_MIN;
-  if (tablespace_flags != USE_FSP_FLAGS)
-  {
-    init_page_size_from_flags(tablespace_flags);
-    initial_page_size= physical_page_size;
-  }
-
-  /* Read just enough to get the tablespace flags */
-  size_t bytes= fread(buf, 1, initial_page_size, fil_in);
-
-  if (bytes != initial_page_size)
-  {
-    fprintf(stderr, "Error: Was not able to read the "
-            "minimum page size of %zu bytes. Bytes read "
-            "was %zu\n", initial_page_size, bytes);
-    return false;
-  }
-
-  /* Read space_id and page offset */
-  cur_space= mach_read_from_4(buf + FIL_PAGE_SPACE_ID);
-  cur_page_num= mach_read_from_4(buf + FIL_PAGE_OFFSET);
-
-  /* Get tablespace flags from the FSP header */
-  uint32_t flags= mach_read_from_4(buf + FSP_HEADER_OFFSET +
-                                   FSP_SPACE_FLAGS);
-
-  if (tablespace_flags != USE_FSP_FLAGS)
-  {
-    if (cur_page_num == 0 && flags != tablespace_flags)
-      fprintf(stderr, "Error: Mismatch between provided tablespace "
-              "flags (0x%x) and file flags (0x%x)\n",
-              tablespace_flags, flags);
-  }
-  else
-  {
-    if (cur_page_num)
-    {
-      fprintf(stderr, "Error: First page of the tablespace file "
-              "should be 0, but encountered page number %" PRIu32 ". "
-              "If you are checking multi file system "
-              "tablespace files, please specify the correct "
-              "tablespace flags using --tablespace-flags option.\n",
-              cur_page_num);
-      return false;
-    }
-    /* Initialize page size parameters based on flags */
-    init_page_size_from_flags(flags);
-    /* Read the rest of the page if it's larger than the minimum size */
-    if (physical_page_size > UNIV_ZIP_SIZE_MIN)
-    {
-      /* Read rest of the page 0 to determine crypt_data */
-      ulint bytes= read_file(buf, true, physical_page_size, fil_in);
-      if (bytes != physical_page_size)
-      {
-        fprintf(stderr, "Error: Was not able to read the rest of the "
-		"page of " ULINTPF " bytes. Bytes read was " ULINTPF "\n",
-		physical_page_size - UNIV_ZIP_SIZE_MIN, bytes);
-        return false;
-      }
-    }
-    tablespace_flags= flags;
-  }
-
-  if (physical_page_size < UNIV_ZIP_SIZE_MIN ||
-      physical_page_size > UNIV_PAGE_SIZE_MAX)
-  {
-    fprintf(stderr, "Error: Invalid page size " ULINTPF
-            " encountered\n", physical_page_size);
-    return false;
-  }
-  return true;
-}
-
 int main(
 	int	argc,
 	char	**argv)
@@ -1670,13 +1551,51 @@ int main(
 			}
 		}
 
-		/* Read and validate page 0 */
-		if (!read_and_validate_page0(fil_in, buf)) {
+		/* Read the minimum page size. */
+		bytes = fread(buf, 1, UNIV_ZIP_SIZE_MIN, fil_in);
+		partial_page_read = true;
+
+		if (bytes != UNIV_ZIP_SIZE_MIN) {
+			fprintf(stderr, "Error: Was not able to read the "
+				"minimum page size ");
+			fprintf(stderr, "of %d bytes.  Bytes read was " ULINTPF "\n",
+				UNIV_ZIP_SIZE_MIN, bytes);
+
 			exit_status = 1;
 			goto my_exit;
 		}
 
-		/* Check if tablespace is encrypted */
+		/* enable variable is_system_tablespace when space_id of given
+		file is zero. Use to skip the checksum verification and rewrite
+		for doublewrite pages. */
+		cur_space = mach_read_from_4(buf + FIL_PAGE_SPACE_ID);
+		cur_page_num = mach_read_from_4(buf + FIL_PAGE_OFFSET);
+
+		/* Determine page size, zip_size and page compression
+		from fsp_flags and encryption metadata from page 0 */
+		init_page_size(buf);
+
+		uint32_t flags = mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + buf);
+
+		if (physical_page_size == UNIV_ZIP_SIZE_MIN) {
+			partial_page_read = false;
+		} else {
+			/* Read rest of the page 0 to determine crypt_data */
+			bytes = read_file(buf, partial_page_read, physical_page_size, fil_in);
+			if (bytes != physical_page_size) {
+				fprintf(stderr, "Error: Was not able to read the "
+					"rest of the page ");
+				fprintf(stderr, "of " ULINTPF " bytes.  Bytes read was " ULINTPF "\n",
+					physical_page_size - UNIV_ZIP_SIZE_MIN, bytes);
+
+				exit_status = 1;
+				goto my_exit;
+			}
+			partial_page_read = false;
+		}
+
+
+		/* Now that we have full page 0 in buffer, check encryption */
 		bool is_encrypted = check_encryption(filename, buf);
 
 		/* Verify page 0 contents. Note that we can't allow
@@ -1687,8 +1606,7 @@ int main(
 			allow_mismatches = 0;
 
 			exit_status = verify_checksum(buf, is_encrypted,
-						      &mismatch_count,
-						      tablespace_flags);
+						      &mismatch_count, flags);
 
 			if (exit_status) {
 				fprintf(stderr, "Error: Page 0 checksum mismatch, can't continue. \n");
@@ -1699,8 +1617,7 @@ int main(
 
 		if ((exit_status = rewrite_checksum(
 					filename, fil_in, buf,
-					&pos, is_encrypted,
-					tablespace_flags))) {
+					&pos, is_encrypted, flags))) {
 			goto my_exit;
 		}
 
@@ -1896,7 +1813,7 @@ unexpected_eof:
 first_non_zero:
 			if (is_system_tablespace) {
 				/* enable when page is double write buffer.*/
-				skip_page = is_page_doublewritebuffer();
+				skip_page = is_page_doublewritebuffer(buf);
 			} else {
 				skip_page = false;
 			}
@@ -1917,16 +1834,13 @@ first_non_zero:
 			    && !is_page_free(xdes, physical_page_size, cur_page_num)
 			    && (exit_status = verify_checksum(
 						buf, is_encrypted,
-						&mismatch_count,
-						tablespace_flags))) {
+						&mismatch_count, flags))) {
 				goto my_exit;
 			}
 
-			if (!is_page_doublewritebuffer() &&
-			    (exit_status = rewrite_checksum(
-				    filename, fil_in, buf,
-				    &pos, is_encrypted,
-				    tablespace_flags))) {
+			if ((exit_status = rewrite_checksum(
+						filename, fil_in, buf,
+						&pos, is_encrypted, flags))) {
 				goto my_exit;
 			}
 
